﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Diagnostics.Contracts;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using MMF.CG.Projection;
using MMF.CG.Utility;
using MMDFileParser.MotionParser;
using SlimDX;
using SlimDX.DXGI;

namespace MMF.CG.Camera.CameraMotion
{
    /// <summary>
    /// VMDファイルを用いて動かすカメラ用のモーションを管理するクラス
    /// This class manage camera moved with vmd moton file.
    /// </summary>
    public class VMDCameraMotionProvider:ICameraMotionProvider
    {
        public static VMDCameraMotionProvider OpenFile(string path)
        {
            return new VMDCameraMotionProvider(MotionData.getMotion(File.OpenRead(path)));
        }

        private List<CameraFrameData> CameraFrames;
        private Stopwatch stopWatch;
        private long lastMillisecound;

        /// <summary>
        /// コンストラクタ
        /// モーションを再生する際は、インスタンス作成の後Startメソッドを呼ぶこと
        /// Constractor
        /// If you play camera motion,you should call Start method after making instance.
        /// </summary>
        /// <param name="cameraMotion">VMDファイルのデータ(VMD parsed data)</param>
        public VMDCameraMotionProvider(MotionData cameraMotion)
        {
            CameraFrames = cameraMotion.CameraFrames.CameraFrames;
            CameraFrames.Sort(new CameraFrameData());
            stopWatch=new Stopwatch();
        }

        /// <summary>
        /// モーションを再生する
        /// Start to play motion
        /// </summary>
        public void Start()
        {
            stopWatch.Start();
        }

        /// <summary>
        /// モーションを停止する
        /// Stop to play motion
        /// </summary>
        public void Stop()
        {
            stopWatch.Stop();
        }

        private void Leap(CameraProvider cp,IProjectionMatrixProvider projection,float frame)
        {
            for (int j = 0; j < CameraFrames.Count - 1; j++)
            {
                if (CameraFrames[j].FrameNumber < frame && CameraFrames[j + 1].FrameNumber >= frame)
                {
                    //フレームが挟まれている時
                    uint frameMargin = CameraFrames[j + 1].FrameNumber - CameraFrames[j].FrameNumber;
                    float progress = (frame - CameraFrames[j].FrameNumber) / (float)frameMargin;
                    float ProgX, ProgY, ProgZ, ProgR, ProgL, ProgP; ;
                    ProgX = CameraFrames[j].Curves[0].Evaluate(progress);
                    ProgY = CameraFrames[j].Curves[1].Evaluate(progress);
                    ProgZ = CameraFrames[j].Curves[2].Evaluate(progress);
                    ProgR = CameraFrames[j].Curves[3].Evaluate(progress);
                    ProgL = CameraFrames[j].Curves[4].Evaluate(progress);
                    ProgP = CameraFrames[j].Curves[5].Evaluate(progress);
                    cp.CameraPosition = CGHelper.ComplementTranslate(CameraFrames[j], CameraFrames[j + 1],
                        new Vector3(ProgX, ProgY, ProgZ));
                    Quaternion rotation = CGHelper.ComplementRotateQuaternion(CameraFrames[j], CameraFrames[j + 1],
                        ProgR);
                    float length = CGHelper.Lerp(CameraFrames[j].Distance, CameraFrames[j + 1].Distance, ProgL);
                    float angle = CGHelper.Lerp(CameraFrames[j].ViewAngle, CameraFrames[j + 1].ViewAngle, ProgP);
                    Vector3 Position2target = Vector3.TransformCoordinate(new Vector3(0, 0, -1),
                        Matrix.RotationQuaternion(rotation));
                    Vector3 TargetPosition = cp.CameraPosition + length*Position2target;
                    cp.CameraLookAt = TargetPosition;
                    projection.Fovy = angle;
                    break;
                }
            }
        }

        void ICameraMotionProvider.UpdateCamera(CameraProvider cp, IProjectionMatrixProvider proj)
        {
            if (lastMillisecound == 0) lastMillisecound = stopWatch.ElapsedMilliseconds;
            else
            {
                long currentMillisecound = stopWatch.ElapsedMilliseconds;
                long elapsed = currentMillisecound - lastMillisecound;
                Leap(cp,proj,elapsed/30f);
            }
        }
    }
}
