﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using MMF.CG.Camera;
using SlimDX;
using SlimDX.Direct3D11;
using SlimDX.DXGI;
using Device = SlimDX.Direct3D11.Device;
using Resource = SlimDX.Direct3D11.Resource;

namespace MMF.CG.DeviceManager
{
    public class ScreenContext:IDisposable
    {
        public ScreenContext(Control owner,RenderContext context,MatrixManager manager)
        {
            this.Context = context;
            Device device = context.DeviceManager.Device;
            SampleDescription sampleDesc = new SampleDescription(1, 0);
            SwapChain = new SwapChain(context.DeviceManager.Factory, device, getSwapChainDescription(owner, sampleDesc));
            //ラスタライザの設定
            //深度ステンシルバッファの初期化
            using (Texture2D depthBuffer = new Texture2D(device, getDepthBufferTexture2DDescription(owner, sampleDesc)))
            {
                DepthTargetView = new DepthStencilView(device, depthBuffer);
            }
            //レンダーターゲットの初期化
            using (Texture2D renderTexture = Resource.FromSwapChain<Texture2D>(SwapChain, 0))
            {
                RenderTargetView = new RenderTargetView(device, renderTexture);
            }
            WorldSpace=new WorldSpace();
            BindedControl = owner;
            MatrixManager = manager;
            PanelObserver=new PanelObserver(owner);
            SetViewport();
        }

        /// <summary>
        /// Gets the render target view.
        /// </summary>
        /// <value>
        /// The render target view.
        /// </value>
        public RenderTargetView RenderTargetView { get; private set; }

        /// <summary>
        /// Gets the depth target view.
        /// </summary>
        /// <value>
        /// The depth target view.
        /// </value>
        public DepthStencilView DepthTargetView { get; private set; }

        /// <summary>
        /// Gets or sets the matrix manager.
        /// </summary>
        /// <value>
        /// The matrix manager.
        /// </value>
        public MatrixManager MatrixManager { get; set; }

        /// <summary>
        /// Gets the swap chain.
        /// </summary>
        /// <value>
        /// The swap chain.
        /// </value>
        public SwapChain SwapChain { get; private set; }

        /// <summary>
        /// Gets the binded control.
        /// </summary>
        /// <value>
        /// The binded control.
        /// </value>
        public Control BindedControl { get; private set; }


        /// <summary>
        /// カメラモーションの挙動を設定するプロパティ
        /// </summary>
        public ICameraMotionProvider CameraMotionProvider { get; set; }


        /// <summary>
        /// このスクリーンに結び付けられているワールド空間
        /// </summary>
        public  WorldSpace WorldSpace { get; set; }

        /// <summary>
        /// ユーザーコントロールを監視するクラス
        /// </summary>
        public PanelObserver PanelObserver { get; private set; }

        /// <summary>
        /// MME特殊パラメータ
        /// object_ssの時に取得するセルフシャドウモード
        /// bool parthfにあたる
        /// </summary>
        public bool IsSelfShadowMode1 { get; private set; }

        /// <summary>
        /// MME特殊パラメータ
        /// bool transpにあたる値
        /// </summary>
        public bool IsEnabledTransparent { get; private set; }

        public void SetViewport()
        {
            Context.DeviceManager.Context.Rasterizer.SetViewports(getViewport());
        }


        public void Resize()
        {
            if (BindedControl.Width == 0 || BindedControl.Height == 0) return; //フォームがフロート状態になった時一瞬だけ来て、デバイスが作れなくなるのでこの時はなしにする。
            if (RenderTargetView != null && !RenderTargetView.Disposed) RenderTargetView.Dispose();
            if (DepthTargetView != null && !DepthTargetView.Disposed) DepthTargetView.Dispose();
            SwapChainDescription desc = SwapChain.Description;
            SwapChain.ResizeBuffers(desc.BufferCount, BindedControl.Width, BindedControl.Height, desc.ModeDescription.Format,
                desc.Flags);
            //深度ステンシルバッファの初期化
            using (Texture2D depthBuffer = new Texture2D(Context.DeviceManager.Device, getDepthBufferTexture2DDescription(BindedControl, desc.SampleDescription)))
            {
                DepthTargetView = new DepthStencilView(Context.DeviceManager.Device, depthBuffer);
            }
            //レンダーターゲットの初期化
            using (Texture2D renderTexture = Resource.FromSwapChain<Texture2D>(SwapChain, 0))
            {
                RenderTargetView = new RenderTargetView(Context.DeviceManager.Device, renderTexture);
            }
        }

        public void MoveCameraByCameraMotionProvider()
        {
            if(CameraMotionProvider==null)return;
            CameraMotionProvider.UpdateCamera(MatrixManager.ViewMatrixManager,MatrixManager.ProjectionMatrixManager);
        }

        /// <summary>
        ///     スワップチェーンの設定を取得します。
        ///     スワップチェーンの設定を変えたい場合は、オーバーライドしてください。
        /// </summary>
        /// <param name="control">適応するコントロールへの参照</param>
        /// <returns>スワップチェーンの設定</returns>
        protected virtual SwapChainDescription getSwapChainDescription(Control control, SampleDescription sampDesc)
        {
            return new SwapChainDescription
            {
                BufferCount = 2,
                Flags = SwapChainFlags.AllowModeSwitch,
                IsWindowed = true,
                ModeDescription = new ModeDescription
                {
                    Format = Format.R8G8B8A8_UNorm,
                    Height = control.Height,
                    Width = control.Width,
                    RefreshRate = new Rational(60, 1)
                },
                OutputHandle = control.Handle,
                SampleDescription = sampDesc,
                SwapEffect = SwapEffect.Discard,
                Usage = Usage.RenderTargetOutput
            };
        }

        /// <summary>
        ///     深度ステンシルバッファの設定を取得します。
        ///     深度ステンシルバッファの設定を変えたい場合はオーバーライドしてください。
        /// </summary>
        /// <param name="control">適用するコントロールへの参照</param>
        /// <returns>深度ステンシルバッファ用のTexture2Dの設定</returns>
        protected virtual Texture2DDescription getDepthBufferTexture2DDescription(Control control, SampleDescription desc)
        {
            return new Texture2DDescription
            {
                ArraySize = 1,
                BindFlags = BindFlags.DepthStencil,
                Format = Format.D32_Float,
                Width = control.Width,
                Height = control.Height,
                MipLevels = 1,
                SampleDescription = desc
            };
        }

        /// <summary>
        ///     ビューポートの内容を取得します。
        /// </summary>
        /// <param name="control">適用するコントロールへの参照</param>
        /// <returns>設定するビューポート</returns>
        protected virtual Viewport getViewport()
        {
            return new Viewport
            {
                Width = BindedControl.Width,
                Height = BindedControl.Height,
                MaxZ = 1
            };
        }

        public void Dispose()
        {
            if (WorldSpace!=null)
            {
                WorldSpace.Dispose();
            }
            if (RenderTargetView != null && !RenderTargetView.Disposed) RenderTargetView.Dispose();
            if (DepthTargetView != null && !DepthTargetView.Disposed) DepthTargetView.Dispose();
            if (SwapChain != null && !SwapChain.Disposed) SwapChain.Dispose();
        }

        public RenderContext Context { get; set; }
    }
}
