﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using MMDD.CG.Model.Sprite;
using MMF.CG.Model;
using MMF.CG.Motion;

namespace MMF.CG.DeviceManager
{
    public class WorldSpace:IDisposable
    {
        private List<IDrawable> drawableResources=new List<IDrawable>();
        private List<IMovable> moveResources=new List<IMovable>();
        private List<IDynamicTexture> dynamicTextures=new List<IDynamicTexture>();
        private List<IGroundShadowDrawable> groundShadowDrawables=new List<IGroundShadowDrawable>();
        private List<IEdgeDrawable> edgeDrawables=new List<IEdgeDrawable>();

        public List<IDrawable> DrawableResources
        {
            get { return drawableResources; }
            private set { drawableResources = value; }
        }

        public List<IMovable> MoveResources
        {
            get { return moveResources; }
            private set { moveResources = value; }
        }

        public List<IDynamicTexture> DynamicTextures
        {
            get { return dynamicTextures; }
            private set { dynamicTextures = value; }
        }

        public List<IGroundShadowDrawable> GroundShadowDrawables
        {
            get { return groundShadowDrawables; }
            private set { groundShadowDrawables = value; }
        }

        public List<IEdgeDrawable> EdgeDrawables
        {
            get { return edgeDrawables; }
            private set { edgeDrawables = value; }
        }

        /// <summary>
        ///     表示するリソースを追加する
        /// </summary>
        /// <param name="drawable"></param>
        public void AddResource(IDrawable drawable)
        {
            drawableResources.Add(drawable);
            if (drawable is IMovable)
            {
                moveResources.Add((IMovable)drawable);
            }
            if (drawable is IEdgeDrawable)
            {
                edgeDrawables.Add((IEdgeDrawable)drawable);
            }
            if (drawable is IGroundShadowDrawable)
            {
                groundShadowDrawables.Add((IGroundShadowDrawable)drawable);
            }
        }

        /// <summary>
        ///     表示するリソースから削除する
        /// </summary>
        /// <param name="drawable"></param>
        public void RemoveResource(IDrawable drawable)
        {
            if (drawableResources.Remove(drawable))
            {
                if (drawable is IMovable)
                {
                    moveResources.Remove((IMovable)drawable);
                }
            }
        }

        /// <summary>
        /// 登録されているすべての描画の必要があるものを描画します
        /// </summary>
        public void DrawAllResources()
        {
            foreach (var edgeDrawable in edgeDrawables)
            {
                if (edgeDrawable.Visiblity) edgeDrawable.DrawEdge();
            }
            foreach (IDrawable d in DrawableResources)
            {
                if (d.Visiblity) d.Draw();
            }
            foreach (var groundShadowDrawable in groundShadowDrawables)
            {
                if (groundShadowDrawable.Visiblity) groundShadowDrawable.DrawGroundShadow();
            }
        }

        /// <summary>
        /// 更新する動的テクスチャを追加します
        /// </summary>
        /// <param name="dtexture">更新する必要のある動的テクスチャ</param>
        public void AddDynamicTexture(IDynamicTexture dtexture)
        {
            dynamicTextures.Add(dtexture);
        }

        /// <summary>
        /// 動的テクスチャを更新対象から除外します
        /// </summary>
        /// <param name="dtexture">もういらない動的テクスチャ</param>
        public void RemoveDynamicTexture(IDynamicTexture dtexture)
        {
            if (dynamicTextures.Contains(dtexture)) dynamicTextures.Remove(dtexture);
        }


        public void Dispose()
        {
            foreach (var drawableResource in DrawableResources)
            {
                if (drawableResource != null) drawableResource.Dispose();
            }
            foreach (var dynamicTexture in DynamicTextures)
            {
                if(dynamicTextures!=null)dynamicTexture.Dispose();
            }
        }

        public void UpdateAllDynamicTexture()
        {
            foreach (var dynamicTexture in DynamicTextures)
            {
                dynamicTexture.UpdateTexture();
            }
        }
    }
}
