﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Threading.Tasks;
using System.Windows.Forms;
using MMF.CG.Camera;
using MMF.CG.Light;
using MMF.CG.MMEEffect;
using MMF.CG.MMEEffect.VariableSubscriber.ConstantSubscriber.ConstantBufferManager;
using MMF.CG.Model.MMD.Morph;
using MMF.CG.Motion;
using MMF.CG.Skinning;
using MMDFileParser.PMXModelParser;
using SlimDX;
using SlimDX.Direct3D11;
using SlimDX.DXGI;
using Device = SlimDX.Direct3D11.Device;

namespace MMF.CG.Model.MMD
{
    public class MMDModel : IMovable,IEdgeDrawable,IGroundShadowDrawable
    {
        public MMDModel(ModelData modeldata, ISubresourceLoader subResourceLoader, string filename)
        {
            Model = modeldata;
            SubresourceLoader = subResourceLoader;
            Transformer = new BasicTransformer();
            SubsetManager = new BasicSubsetManager(this);
            //PhongConstantBuffer = new PhongShadingConstantBufferManager();
            ToonManager = new BasicToonTextureManager();
            SelfShadowColor = new Vector4(0, 0, 0, 1);
            GroundShadowColor = new Vector4(0, 0, 0, 1);
            FileName = filename;
            Visiblity = true;
        }

        private MMDModel()
        {
            
        }

        /// <summary>
        /// テクスチャなどのリソースのパスを解決するインターフェース
        /// </summary>
        public ISubresourceLoader SubresourceLoader { get; private set; }

        /// <summary>
        ///     読み込み対象のモデル
        /// </summary>
        public ModelData Model { get; private set; }


        /// <summary>
        ///     バッファ管理インターフェース
        /// </summary>
        public IBufferManager BufferManager { get; private set; }

        /// <summary>
        /// サブセット管理インターフェース
        /// </summary>
        public ISubsetManager SubsetManager { get; private set; }

        public IMMDMorphManager Morphmanager { get; private set; }


        /// <summary>
        /// 描画に利用するMMEエフェクト
        /// </summary>
        public MMEEffectManager Effect { get; private set; }



       // public PhongShadingConstantBufferManager PhongConstantBuffer { get; private set; }

        /// <summary>
        /// トゥーンを管理してくれるインターフェース
        /// </summary>
        public IToonTextureManager ToonManager { get; private set; }

        /// <summary>
        /// レンダーコンテキスト
        /// </summary>
        public RenderContext RenderContext { get; private set; }

        private EffectPass ZPlotPass { get; set; }

        /// <summary>
        ///     初期化判定
        /// </summary>
        private bool IsInitialized { get; set; }

        #region IDrawableインターフェース実装部

        private Vector4 groundShadowColor;
        //private PhongShadingConstantBufferInputLayout phongBuffer;
        private Vector4 selfShadowColor;
        public bool Visiblity { get; set; }
        public string FileName { get; set; }

        public void LoadFromEffectFile(string filePath)
        {
            Effect = MMEEffectManager.Load(filePath, this, RenderContext, SubresourceLoader);
        }

        public int SubsetCount
        {
            get
            {
                return SubsetManager.SubsetCount;
            }
        }

        public int VertexCount
        {
            get
            {
                return BufferManager.VerticiesCount;
            }
        }

        public virtual void Draw()
        {
            PreModelDraw(RenderContext.DeviceManager.Device);
            UpdateConstantByModel();
            Effect.ApplyAllMatrixVariables();
            Skinning.ApplyEffect(Effect.EffectFile);
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.SetVertexBuffers(0,
                new VertexBufferBinding(BufferManager.VertexBuffer, BasicInputLayout.SizeInBytes, 0));
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.SetIndexBuffer(
                BufferManager.IndexBuffer, Format.R16_UInt, 0);
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.InputLayout = BufferManager.VertexLayout;
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.PrimitiveTopology =
                PrimitiveTopology.TriangleList;

            SubsetManager.DrawAll();

            PostModelDraw(RenderContext.DeviceManager.Device);
        }


        public void DrawEdge()
        {
            UpdateConstantByModel();
            Effect.ApplyAllMatrixVariables();
            Skinning.ApplyEffect(Effect.EffectFile);
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.SetVertexBuffers(0,
                new VertexBufferBinding(BufferManager.VertexBuffer, BasicInputLayout.SizeInBytes, 0));
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.SetIndexBuffer(
                BufferManager.IndexBuffer, Format.R16_UInt, 0);
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.InputLayout = BufferManager.VertexLayout;
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.PrimitiveTopology =
                PrimitiveTopology.TriangleList;
            SubsetManager.DrawEdges();
        }

        public void DrawGroundShadow()
        {
            UpdateConstantByModel();
            Effect.ApplyAllMatrixVariables();
            Skinning.ApplyEffect(Effect.EffectFile);
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.SetVertexBuffers(0,
                new VertexBufferBinding(BufferManager.VertexBuffer, BasicInputLayout.SizeInBytes, 0));
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.SetIndexBuffer(
                BufferManager.IndexBuffer, Format.R16_UInt, 0);
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.InputLayout = BufferManager.VertexLayout;
            RenderContext.DeviceManager.Device.ImmediateContext.InputAssembler.PrimitiveTopology =
                PrimitiveTopology.TriangleList;
            SubsetManager.DrawGroundShadow();
        }

        public void Update()
        {
            if(BufferManager!=null)BufferManager.RecreateVerticies();
            Morphmanager.UpdateFrame();
            Skinning.UpdateSkinning(Morphmanager);
            foreach (var pmxSubset in SubsetManager.Subsets)
            {
                pmxSubset.MaterialInfo.UpdateMaterials();
            }
        }

        /// <summary>
        ///     1000/30msごとに呼び出され、フレームを更新します。
        /// </summary>
        public virtual void ApplyMove()
        {
            //PreApplyMove();
            
            //PostApplyMove();
        }

        public void GetFileName()
        {

        }

        public void Load(RenderContext renderContext)
        {
            Stopwatch sw = new Stopwatch();
            sw.Start();
            RenderContext = renderContext;
            Device device = RenderContext.DeviceManager.Device;
            if (!IsInitialized)
            {
//モデルの読み込みなどはリセットしない
                //シェーダーの読み込み
                Effect = InitializeEffect();
                ToonManager.Initialize(device, SubresourceLoader);
                SubsetManager.Initialze(Model, device, Effect, SubresourceLoader,
                    ToonManager);
                //定数バッファ
                //PhongConstantBuffer.Initialize(device, Effect.EffectFile,
                //    PhongShadingConstantBufferInputLayout.SizeInBytes, new PhongShadingConstantBufferInputLayout());
                ZPlotPass = Effect.EffectFile.GetTechniqueByIndex(1).GetPassByIndex(0);
                InitializeBuffers(device);
                Skinning = InitializeSkinning();
                Morphmanager=new MMDMorphManager(this);
                IsInitialized = true;
            }
            MotionManager = InitializeMotionManager();
            InitializeOther(device);
            sw.Stop();
            Trace.WriteLine(sw.ElapsedMilliseconds + "ms");
        }


        public Task LoadAsync(RenderContext renderContext)
        {
            Task task=new Task(() => Load(renderContext));
            task.Start();
            return task;
        }

        /// <summary>
        ///     現状の定数バッファを更新します
        /// </summary>
        private void UpdateConstantByModel()
        {
            //phongBuffer.LightPointPosition = new Vector4(0, 0, -100, 0);
            //phongBuffer.ViewPointPosition = new Vector4(RenderContext.MatrixManager.ViewMatrixManager.CameraPosition, 0);
            //PhongConstantBuffer.ApplyToEffect(phongBuffer);
        }

        #endregion

        #region 初期化系

        public void LoadEffect(string effectFile)
        {
            if (Effect != null) Effect.Dispose();
            if (string.IsNullOrEmpty(effectFile))
            {
                //TODO デフォルトエフェクトのデバッグ用の突貫工事
                Effect = MMEEffectManager.Load(@"..\..\..\MikuMikuFlex\Shader\DefaultShader.fx", this, RenderContext, SubresourceLoader);
            }
            else
            {
                Effect = MMEEffectManager.Load(effectFile, this, RenderContext, SubresourceLoader);
            }
            SubsetManager.ResetEffect(Effect);
        }

        protected virtual void PreModelDraw(Device device)
        {
        }

        protected virtual void PostModelDraw(Device device)
        {
        }

        protected virtual void PreApplyMove()
        {
        }

        protected virtual void PostApplyMove()
        {
        }

        protected virtual ISkinningProvider InitializeSkinning()
        {
            Skeleton skel=new Skeleton(Model);
            return skel;
        }

        protected virtual IMotionManager InitializeMotionManager()
        {
            BasicMotionManager mm = new BasicMotionManager(RenderContext);
            mm.Initialize(Model,Morphmanager,Skinning,BufferManager);
            ((Skeleton)Skinning).KinematicsProviders.Insert(0,mm);
            return mm;
        }

        protected virtual MMEEffectManager InitializeEffect()
        {
            return MMEEffectManager.LoadFromResource(@"MMF.Shader.DefaultShader.fx", this, RenderContext, SubresourceLoader);
        }

        protected virtual void InitializeOther(Device device)
        {
        }

        protected virtual void InitializeBuffers(Device device)
        {
            BufferManager = new MMDModelBufferManager();
            BufferManager.Initialize(Model, device, Effect.EffectFile);
        }

        #endregion

        /// <summary>
        /// セルフシャドウの色
        /// </summary>
        public Vector4 SelfShadowColor
        {
            get { return selfShadowColor; }
            set
            {
                selfShadowColor = value;
            }
        }

        /// <summary>
        /// 地面影の色
        /// </summary>
        public Vector4 GroundShadowColor
        {
            get { return groundShadowColor; }
            set
            {
                groundShadowColor = value;
            }
        }

        /// <summary>
        ///     モデルの位置や回転量を管理するインターフェース
        /// </summary>
        public ITransformer Transformer { get; private set; }

        public virtual void Dispose()
        {
            Effect.Dispose();
            BufferManager.Dispose();
            //PhongConstantBuffer.Dispose();
            ToonManager.Dispose();
            SubsetManager.Dispose();
        }

        /// <summary>
        ///     モーション管理クラス
        /// </summary>
        public IMotionManager MotionManager { get; private set; }

        /// <summary>
        ///     スキニングのインターフェース
        /// </summary>
        public ISkinningProvider Skinning { get; private set; }

        #region 静的初期化便利メソッド

        #region FromFile
        /// <summary>
        ///     モデルファイルを開く
        /// </summary>
        /// <param name="filePath">PMXのファイルパス、テクスチャは同じフォルダから読み込まれる</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static MMDModel FromFile(string filePath)
        {
            string folder = Path.GetDirectoryName(filePath);
            return FromFile(filePath, folder);
        }

        /// <summary>
        ///     モデルファイルを開く
        /// </summary>
        /// <param name="filePath">PMXのファイルパス</param>
        /// <param name="textureFolder">テクスチャを読み込むフォルダ</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static MMDModel FromFile(string filePath, string textureFolder)
        {
            return FromFile(filePath, new BasicSubresourceLoader(textureFolder));
        }

        /// <summary>
        /// モデルファイルを開く
        /// </summary>
        /// <param name="filePath">PMXのファイルパス</param>
        /// <param name="loader">テクスチャのパス解決インターフェース</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static MMDModel FromFile(string filePath, ISubresourceLoader loader)
        {
            using (FileStream fs = File.OpenRead(filePath))
            {
                return new MMDModel(ModelData.GetModel(fs),loader, Path.GetFileName(filePath));
            }
        }
        #endregion

        #region OpenLoad

        /// <summary>
        /// 開いて初期化
        /// </summary>
        /// <param name="filePath">PMXのファイルパス</param>
        /// <param name="context">レンダリングコンテキスト</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static MMDModel OpenLoad(string filePath,RenderContext context)
        {
            MMDModel model = FromFile(filePath);
            model.Load(context);
            return model;
        }

        /// <summary>
        /// 開いて初期化
        /// </summary>
        /// <param name="filePath">PMXファイルのパス</param>
        /// <param name="textureFolder">テクスチャのフォルダ</param>
        /// <param name="context">レンダリングコンテキスト</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static MMDModel OpenLoad(string filePath, string textureFolder, RenderContext context)
        {
            MMDModel model = FromFile(filePath, textureFolder);
            model.Load(context);
            return model;
        }

        /// <summary>
        /// 開いて初期化
        /// </summary>
        /// <param name="filePath">PMXファイルのパス</param>
        /// <param name="loader">テクスチャなどのパスの解決インターフェース</param>
        /// <param name="context">レンダリングコンテキスト</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static MMDModel OpenLoad(string filePath, ISubresourceLoader loader, RenderContext context)
        {
            MMDModel model = FromFile(filePath, loader);
            model.Load(context);
            return model;
        }

        /// <summary>
        /// 開いて初期化
        /// </summary>
        /// <param name="filePath">PMXのファイルパス</param>
        /// <param name="panel">レンダーパネル</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static MMDModel OpenLoad(string filePath, RenderControl panel)
        {
            MMDModel model = FromFile(filePath);
            model.Load(panel.RenderContext);
            return model;
        }

        /// <summary>
        /// 開いて初期化
        /// </summary>
        /// <param name="filePath">PMXファイルのパス</param>
        /// <param name="textureFolder">テクスチャのフォルダ</param>
        /// <param name="panel">レンダーパネル</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static MMDModel OpenLoad(string filePath, string textureFolder, RenderControl panel)
        {
            MMDModel model = FromFile(filePath, textureFolder);
            model.Load(panel.RenderContext);
            return model;
        }

        /// <summary>
        /// 開いて初期化
        /// </summary>
        /// <param name="filePath">PMXファイルのパス</param>
        /// <param name="loader">テクスチャなどのパスの解決インターフェース</param>
        /// <param name="panel">レンダリングコンテキスト</param>
        /// <returns></returns>
        public static MMDModel OpenLoad(string filePath, ISubresourceLoader loader, RenderControl panel)
        {
            MMDModel model = FromFile(filePath, loader);
            model.Load(panel.RenderContext);
            return model;
        }
        #endregion

        #region OpenLoadAsync
        /// <summary>
        /// 開いて初期化(非同期)
        /// </summary>
        /// <param name="filePath">PMXのファイルパス</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static Task<MMDModel> OpenLoadAsync(string filePath, RenderContext context)
        {
            Task<MMDModel> task=new Task<MMDModel>(() =>
            {
                MMDModel model = FromFile(filePath);
                model.Load(context);
                return model;
            });
            task.Start();
            return task;
        }

        /// <summary>
        /// 開いて初期化(非同期)
        /// </summary>
        /// <param name="filePath">PMXファイルのパス</param>
        /// <param name="textureFolder">テクスチャのフォルダ</param>
        /// <param name="context">レンダリングコンテキスト</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static Task<MMDModel> OpenLoadAsync(string filePath, string textureFolder, RenderContext context)
        {
            Task<MMDModel> task = new Task<MMDModel>(() =>
            {
                MMDModel model = FromFile(filePath, textureFolder);
                model.Load(context);
                return model;
            });
            task.Start();
            return task;
        }

        /// <summary>
        /// 開いて初期化(非同期)
        /// </summary>
        /// <param name="filePath">PMXファイルのパス</param>
        /// <param name="loader">テクスチャなどのパスの解決インターフェース</param>
        /// <param name="context">レンダリングコンテキスト</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static Task<MMDModel> OpenLoadAsync(string filePath, ISubresourceLoader loader, RenderContext context)
        {
            Task<MMDModel> task = new Task<MMDModel>(() =>
            {
                MMDModel model = FromFile(filePath, loader);
                model.Load(context);
                return model;
            });
            task.Start();
            return task;

        }

        /// <summary>
        /// 開いて初期化(非同期)
        /// </summary>
        /// <param name="filePath">PMXのファイルパス</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static Task<MMDModel> OpenLoadAsync(string filePath, RenderControl panel)
        {
            Task<MMDModel> task = new Task<MMDModel>(() =>
            {
                MMDModel model = FromFile(filePath);
                model.Load(panel.RenderContext);
                return model;
            });
            task.Start();
            return task;
        }

        /// <summary>
        /// 開いて初期化(非同期)
        /// </summary>
        /// <param name="filePath">PMXファイルのパス</param>
        /// <param name="textureFolder">テクスチャのフォルダ</param>
        /// <param name="panel">レンダーパネル</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static Task<MMDModel> OpenLoadAsync(string filePath, string textureFolder, RenderControl panel)
        {
            Task<MMDModel> task = new Task<MMDModel>(() =>
            {
                MMDModel model = FromFile(filePath, textureFolder);
                model.Load(panel.RenderContext);
                return model;
            });
            task.Start();
            return task;

        }

        /// <summary>
        /// 開いて初期化(非同期)
        /// </summary>
        /// <param name="filePath">PMXファイルのパス</param>
        /// <param name="loader">テクスチャなどのパスの解決インターフェース</param>
        /// <param name="panel">レンダーパネル</param>
        /// <returns>MMDModelのインスタンス</returns>
        public static Task<MMDModel> OpenLoadAsync(string filePath, ISubresourceLoader loader, RenderControl panel)
        {
            Task<MMDModel> task = new Task<MMDModel>(() =>
            {
                MMDModel model = FromFile(filePath, loader);
                model.Load(panel.RenderContext);
                return model;
            });
            task.Start();
            return task;
        }
        #endregion


        #endregion


    }
}