﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using MMF.CG.Model;
using MMF.CG.Model.MMD;
using MMF.CG.Model.MMD.Morph;
using MMF.CG.Skinning;
using MMDFileParser.PMXModelParser;

namespace MMF.CG.Motion
{
    /// <summary>
    ///     モーション管理クラス
    /// </summary>
    public class BasicMotionManager : IMotionManager
    {
        private readonly RenderContext context;

        private Stopwatch motionTimer;

        /// <summary>
        ///     スキニングに利用するインターフェース
        /// </summary>
        private ISkinningProvider skinningProvider;

        private IMMDMorphManager morphManager;

        public BasicMotionManager(RenderContext context)
        {
            this.context = context;
        }

        private long lastTime { get; set; }

        /// <summary>
        ///     現在再生中のモーションのindex
        /// </summary>
        public IMotionProvider CurrentMotionProvider { get; set; }

        /// <summary>
        ///     現在再生中のモーションのフレーム位置(秒単位)を取得します。
        /// </summary>
        public float CurrentFrame
        {
            get
            {
                if (CurrentMotionProvider == null) return float.NaN;
                return CurrentMotionProvider.CurrentFrame/context.Timer.MotionFramePerSecond;
            }
        }

        /// <summary>
        ///     前回のフレームからどれだけ時間がかかったか(秒単位)取得します
        /// </summary>
        public float ElapsedTime { get; private set; }

        /// <summary>
        ///     初期化
        /// </summary>
        /// <param name="skinning">スキニングに利用するインターフェース</param>
        public void Initialize(ModelData model,IMMDMorphManager morph,ISkinningProvider skinning,IBufferManager bufferManager)
        {
            skinningProvider = skinning;
            motionTimer=new Stopwatch();
            motionTimer.Start();
            this.morphManager = morph;
            SubscribedMotions = new List<MotionNameAndIndexData>();
        }

        /// <summary>
        ///     現在のモーションを進める
        /// </summary>
        public bool UpdateTransform(ISkinningProvider prov, IMMDMorphManager morph)
        {
            if (lastTime == 0)
            {
                lastTime = motionTimer.ElapsedMilliseconds;
            }
            else
            {
                long currentTime = motionTimer.ElapsedMilliseconds;
                ElapsedTime = currentTime - lastTime;
                if (CurrentMotionProvider != null) CurrentMotionProvider.Tick();
                lastTime = currentTime;
            }
            return true;
        }


        /// <summary>
        ///     モーションをファイルから追加する
        /// </summary>
        /// <param name="filePath">ファイルパス</param>
        /// <returns>モーションのindex</returns>
        public IMotionProvider AddMotionFromFile(string filePath,bool ignoreParent)
        {
            IMotionProvider motion = new MMDMotion(filePath, context,this,morphManager,ignoreParent);
            motion.LoadMotion(skinningProvider);
            motion.MotionFinished += motion_MotionFinished;
            SubscribedMotions.Add(new MotionNameAndIndexData(motion, filePath));
            if (MotionListUpdated != null) MotionListUpdated(this, new EventArgs());
            return motion;
        }

        /// <summary>
        ///     指定したモーションを再生する
        /// </summary>
        /// <param name="id">モーションのid</param>
        /// <param name="startFrame">最初のフレーム</param>
        /// <param name="setting">終了後の挙動</param>
        public void ApplyMotion(IMotionProvider motionProvider, int startFrame = 0,
            ActionAfterMotion setting = ActionAfterMotion.Nothing)
        {
            if (CurrentMotionProvider != null) CurrentMotionProvider.Stop();
            motionProvider.Start(startFrame, setting);
            CurrentMotionProvider = motionProvider;
        }

        /// <summary>
        ///     モーション再生を終了する
        /// </summary>
        public void StopMotion()
        {
            CurrentMotionProvider.Stop();
        }

        /// <summary>
        ///     モーションが終了したことを通知する
        /// </summary>
        public event EventHandler<ActionAfterMotion> MotionFinished;

        /// <summary>
        ///     登録されたモーションのリスト
        /// </summary>
        public List<MotionNameAndIndexData> SubscribedMotions { get; private set; }

        /// <summary>
        ///     モーションのリストが更新されたことを通知する
        /// </summary>
        public event EventHandler MotionListUpdated;

        public IMotionProvider AddMotionFromStream(string fileName, Stream stream,bool ignoreParent)
        {
            IMotionProvider motion = new MMDMotion(stream, context,this,morphManager,ignoreParent);
            motion.MotionFinished += motion_MotionFinished;
            SubscribedMotions.Add(new MotionNameAndIndexData(motion, fileName));
            if (MotionListUpdated != null)MotionListUpdated(this, new EventArgs());
            return motion;
        }

        private void motion_MotionFinished(object owner, ActionAfterMotion obj)
        {
            if (MotionFinished != null) MotionFinished(this, obj);
        }
    }
}