﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using MMF.CG.Model.MMD;
using MMF.CG.Model.MMD.Morph;
using MMF.CG.Skinning;
using MMDFileParser.MotionParser;

namespace MMF.CG.Motion
{
    /// <summary>
    /// モーションを管理するクラス
    /// </summary>
    public class MMDMotion : IMotionProvider
    {
        /// <summary>
        ///     動かさなくては行けないボーンのリストを保持するリスト
        /// </summary>
        private readonly List<BoneMotion> boneMotion = new List<BoneMotion>();

        private readonly List<FaceMotion> faceMotion=new List<FaceMotion>(); 

        /// <summary>
        ///     モーションデータ
        /// </summary>
        public MotionData Motion { get; private set; }

        /// <summary>
        ///     動かすのが１回だけなボーンを保持するリスト
        /// </summary>
        private readonly List<BoneMotion> onceMotion = new List<BoneMotion>();

        private readonly List<FaceMotion> onceFaceMotion = new List<FaceMotion>(); 


        /// <summary>
        ///     モーション再生終了後の挙動
        /// </summary>
        private ActionAfterMotion afterMotion = ActionAfterMotion.Nothing;

        public event EventHandler<EventArgs> FrameTicked;

        /// <summary>
        ///     このモーションが再生中かどうか
        /// </summary>
        private bool isPlaying;

        private bool isloaded;

        /// <summary>
        ///     スキニングをするインターフェース
        /// </summary>
        private ISkinningProvider skinningProvider;

        private RenderContext context;
        private IMMDMorphManager morphManager;
        private BasicMotionManager motionManager;
        private bool ignoreParent;

        public MMDMotion(string filePath,RenderContext context,BasicMotionManager motionManager,IMMDMorphManager morphManager,bool ignoreParent)
        {
            this.ignoreParent = ignoreParent;
            this.motionManager = motionManager;
            this.context = context;
            FileStream fs = new FileStream(filePath, FileMode.Open);
            Motion = MotionData.getMotion(fs);
            fs.Close();
            this.morphManager = morphManager;
        }

        public MMDMotion(Stream fs,RenderContext context,BasicMotionManager motionManager,IMMDMorphManager morphManager,bool ignoreParent)
        {
            this.ignoreParent = ignoreParent;
            this.motionManager = motionManager;
            this.context = context;
            Motion = MotionData.getMotion(fs);
            this.morphManager = morphManager;
        }

        /// <summary>
        ///     現在のフレーム
        /// </summary>
        public float CurrentFrame { get; set; }

        /// <summary>
        ///     このモーションにおける最終フレーム
        /// </summary>
        public int FinalFrame { get; private set; }

        /// <summary>
        ///     モーションが終了したことを表すイベント
        /// </summary>
        public event EventHandler<ActionAfterMotion> MotionFinished;

        /// <summary>
        ///     モデルからボーンの情報を読み出し、モーションを当てられる状態にします。
        /// </summary>
        /// <param name="skinning">利用するスキニング管理インターフェース</param>
        public void LoadMotion(ISkinningProvider skinning)
        {
            this.skinningProvider = skinning;
            //フレームを読み込む(ボーンモーション)
            for (int i = 0; i < Motion.BoneFrames.BoneFrameCount; i++)
            {
                //現在読み込まなければいけないボーンフレーム
                BoneFrameData boneFrame = Motion.BoneFrames.boneFrameList[i];
                bool isLoadedBone = false;
                if (ignoreParent&&boneFrame.BoneName.Equals("全ての親"))continue;
                int boneId = getBoneIndex(boneFrame.BoneName);
                if (boneId >= 0)
                {
                    foreach (BoneMotion item in boneMotion) //すでに同じボーンで読み込まれているものがないか検査
                    {
                        //すでにそのボーンで読み込まれているフレームがあった場合はそのフレームのリストに追加する。
                        if (item.BoneIndex == boneId)
                        {
                            item.Frames.Add(boneFrame);
                            isLoadedBone = true;
                            break;
                        }
                    }
                    if (!isLoadedBone) //上のループで読み込まれなかった場合。つまり、今までに同じボーンのフレームが読み込まれてなかった場合は、新たにリストにインスタンスを作成し、加える。
                    {
                        BoneMotion bm = new BoneMotion(boneId);
                        bm.Frames.Add(boneFrame);
                        boneMotion.Add(bm);
                    }
                }
            }

            for (int i = 0; i < Motion.FaceFrames.FaceFrameListCount; i++)
            {
                FaceFrameData faceFrame = Motion.FaceFrames.FaceFrames[i];
                bool isLoadedFace = false;
                string morphName = faceFrame.Name;
                var loaded = (from face in faceMotion where face.Morphname == morphName select face).ToArray();
                if (loaded.Any())
                {
                    loaded[0].Frames.Add(faceFrame);
                    isLoadedFace = true;
                }
                if (!isLoadedFace)
                {
                    FaceMotion fm=new FaceMotion(morphName);
                    fm.Frames.Add(faceFrame);
                    faceMotion.Add(fm);
                }
            }
            foreach (FaceMotion motion in faceMotion)
            {
                motion.Frames.Sort(new FaceFrameData());
            }

            //フレームのソートと最終フレームの検出
            foreach (BoneMotion item in boneMotion)
            {
                item.Frames.Sort(new BoneFrameData());
                FinalFrame = Math.Max((int) item.Frames[item.Frames.Count - 1].FrameNumber, FinalFrame);
            }
            //フレームの整理。1フレームしかないボーンモーションを最初からあてがう。
            for (int i = 0; i < boneMotion.Count; i++)
            {
                if (boneMotion[i].Frames.Count == 1)
                {
                    skinning.Bone[boneMotion[i].BoneIndex].Rotation = boneMotion[i].Frames[0].BoneRotatingQuaternion;
                    skinning.Bone[boneMotion[i].BoneIndex].Translation = boneMotion[i].Frames[0].BonePosition;
                    onceMotion.Add(boneMotion[i]);
                    boneMotion.RemoveAt(i);
                }
            }
            isloaded = true;
        }

        /// <summary>
        ///     モーションを進める
        /// </summary>
        public void Tick()
        {
            if (isPlaying)
            {
                if (CurrentFrame >= FinalFrame)
                {
                    if (MotionFinished != null) MotionFinished(this, afterMotion);
                    if (afterMotion == ActionAfterMotion.Replay) CurrentFrame = 0;
                }
                ApplyToModelMovingMatrix(CurrentFrame);
                CurrentFrame += (float)motionManager.ElapsedTime/1000f*context.Timer.MotionFramePerSecond;
                if (CurrentFrame >= FinalFrame)
                    CurrentFrame = FinalFrame;
                if(FrameTicked!=null)FrameTicked(this,new EventArgs());
            }
            else
            {
                ApplyToModelMovingMatrix(CurrentFrame);
            }
        }

        /// <summary>
        ///     モーションを再生する
        /// </summary>
        /// <param name="frame">再生フレーム番号</param>
        /// <param name="action"></param>
        public void Start(float frame, ActionAfterMotion action)
        {
            if (frame > FinalFrame)
            {
                throw new InvalidOperationException("最終フレームを超えた場所から再生を求められました。");
            }
            CurrentFrame = frame;
            isPlaying = true;
            afterMotion = action;
        }

        /// <summary>
        ///     モーションの更新を停止する
        /// </summary>
        public void Stop()
        {
            isPlaying = false;
        }

        /// <summary>
        ///     ロード済みかどうか
        /// </summary>
        public bool isLoaded
        {
            get { return isloaded; }
        }

        /// <summary>
        ///     スキニング管理クラスに対して回転量と移動量を割り当てる
        /// </summary>
        /// <param name="frame">割り当てるフレーム番号</param>
        public void ApplyToModelMovingMatrix(float frame)
        {
            if (Math.Abs(frame) < 1.0e-3) //フレームがひとつの場合のモーションを当てる
            {
                foreach (BoneMotion item in onceMotion)
                {
                    skinningProvider.Bone[item.BoneIndex].Rotation = item.Frames[0].BoneRotatingQuaternion;
                    skinningProvider.Bone[item.BoneIndex].Translation = item.Frames[0].BonePosition;
                }
            }
            else
            {
                Parallel.For(0, boneMotion.Count, (i) =>
                {
                    BoneMotion bm = boneMotion[i];
                    bm.Lerp(skinningProvider.Bone[bm.BoneIndex], frame);
                });
            }
            morphManager.ApplyMorphProgress(frame,faceMotion);
        }


        /// <summary>
        ///     ボーン名からボーンのインデックスを取得します。存在しない場合-1を返します。
        /// </summary>
        /// <param name="boneName"></param>
        /// <returns></returns>
        private int getBoneIndex(string boneName)
        {
            for (int i = 0; i < skinningProvider.Bone.Length; i++)
            {
                if (skinningProvider.Bone[i].BoneName == boneName)
                {
                    return i;
                }
            }
            return -1;
        }
    }
}