﻿using System.Collections.Generic;
using System.Linq;
using BulletSharp;
using SlimDX;
using IDisposable = System.IDisposable;

/// 物理演算Bulletのラッパークラス

namespace MMF.CG.Physics
{
    /// <summary>
    /// 物理演算のラッパークラス
    /// </summary>
    public class BulletPhysics : IDisposable
    {
        private readonly DefaultCollisionConfiguration collisionConfiguration;
        private readonly AlignedCollisionShapeArray collisionShapes = new AlignedCollisionShapeArray();
        private readonly CollisionDispatcher dispatcher;
        public DiscreteDynamicsWorld dynamicsWorld;
        private readonly BroadphaseInterface overlappingPairCache;
        private readonly SequentialImpulseConstraintSolver solver;

        public readonly Dictionary<RigidBody,CollisionShape> ShapeDictionary=new Dictionary<RigidBody, CollisionShape>(); 

        private bool IsDisposed = false;

        public BulletPhysics(Vector3 gravity)
        {
            collisionConfiguration = new DefaultCollisionConfiguration();
            dispatcher = new CollisionDispatcher(collisionConfiguration);
            overlappingPairCache = new DbvtBroadphase();
            solver = new SequentialImpulseConstraintSolver();
            dynamicsWorld = new DiscreteDynamicsWorld(dispatcher, overlappingPairCache, solver, collisionConfiguration);
            dynamicsWorld.Gravity = gravity;
        }

        public void Dispose()
        {
            if(IsDisposed)return;
                for (int i = dynamicsWorld.NumConstraints - 1; i >= 0; --i)
                {
                    TypedConstraint constraint = dynamicsWorld.GetConstraint(i);
                    dynamicsWorld.RemoveConstraint(constraint);
                    constraint.Dispose();
                }
                for (int i = dynamicsWorld.NumCollisionObjects - 1; i >= 0; --i)
                {
                    CollisionObject obj = dynamicsWorld.CollisionObjectArray[i];
                    RigidBody body = RigidBody.Upcast(obj);
                    if (body != null && body.MotionState != null) body.MotionState.Dispose();
                    dynamicsWorld.RemoveCollisionObject(obj);
                    obj.Dispose();
                }
            for (int i = 0; i < collisionShapes.Count; ++i)
            {
                CollisionShape shape = collisionShapes[i];
                collisionShapes[i] = null;
                shape.Dispose();
            }
            collisionShapes.Dispose();
            dynamicsWorld.Dispose();
            solver.Dispose();
            overlappingPairCache.Dispose();
            dispatcher.Dispose();
            collisionConfiguration.Dispose();
            IsDisposed = true;
        }

        public RigidBody CreatePlane(float planeConstant, Matrix world, float restitution = 1, float friction = 0, float linear_damp = 0, float angular_damp = 0, CollisionFilterGroups groups = CollisionFilterGroups.DefaultFilter, CollisionFilterGroups mask = CollisionFilterGroups.AllFilter)
        {
            CollisionShape shape = new StaticPlaneShape(new Vector3(0, 1, 0), planeConstant);
            return CreateShape(shape, world, 0, restitution, friction, linear_damp, angular_damp, false, groups, mask);
        }

        private RigidBody CreateShape(CollisionShape shape, Matrix world, float mass, float restitution,
            float friction, float linear_damp, float angular_damp, bool kinematic, CollisionFilterGroups group,
            CollisionFilterGroups mask)
        {
            collisionShapes.Add(shape);
            Vector3 localInertia = new Vector3(0, 0, 0);
            if (mass != 0) shape.CalculateLocalInertia(mass, out localInertia);
            DefaultMotionState motionState = new DefaultMotionState(world);
            RigidBodyConstructionInfo rbInfo = new RigidBodyConstructionInfo(mass, motionState, shape, localInertia);
            RigidBody body = new RigidBody(rbInfo);
            body.Restitution = restitution;
            body.Friction = friction;
            body.SetDamping(linear_damp, angular_damp);
            float linearDamp = body.LinearDamping;
            float angularDamp = body.AngularDamping;
            if (kinematic)
            {
                body.CollisionFlags = body.CollisionFlags | CollisionFlags.KinematicObject;
            }
			body.ActivationState = ActivationState.DisableDeactivation;
			dynamicsWorld.AddRigidBody(body, group, mask);
            ShapeDictionary.Add(body,shape);
            return body;
        }


        public void RemoveRigidBody(RigidBody rigidBody)
        {
            dynamicsWorld.RemoveRigidBody(rigidBody);
        }

        /// <summary>
        ///     箱剛体を作成。幅、高さ、深さの定義はBullet仕様なので注意。剛体中心から端までの距離、すなわち辺の長さの1/2で表す。
        /// </summary>
        /// <param name="width">X方向の幅</param>
        /// <param name="height">Y方向の高さ</param>
        /// <param name="depth">Z方向の深さ</param>
        /// <param name="world">初期配置を示すワールド変換行列</param>
        /// <param name="mass">質量。0にすると動かないstatic剛体になる。</param>
        /// <param name="restitution">反発係数</param>
        /// <param name="friction">摩擦係数</param>
        /// <param name="linear_damp">移動減衰係数</param>
        /// <param name="angular_damp">回転減衰係数</param>
        /// <param name="kinematic">物理演算の影響を受けないKinematic剛体</param>
        /// <param name="group">グループNo.</param>
        /// <param name="mask">衝突するグループNo.</param>
        /// <returns></returns>
        public RigidBody CreateBox(float width, float height, float depth, Matrix world, float mass = 0,
            float restitution = 0,
            float friction = 0.5f, float linear_damp = 0, float angular_damp = 0, bool kinematic = false,
            CollisionFilterGroups group = CollisionFilterGroups.DefaultFilter,
            CollisionFilterGroups mask = CollisionFilterGroups.AllFilter
            )
        {
            CollisionShape shape = new BoxShape(width, height, depth);
            if (kinematic) mass = 0;
            return CreateShape(shape, world, mass, restitution, friction, linear_damp, angular_damp, kinematic, group,
                mask);
        }

        /// <summary>
        ///     球剛体を作成
        /// </summary>
        /// <param name="radius">半径</param>
        /// <param name="world">初期配置を示すワールド変換行列</param>
        /// <param name="mass">質量。0にすると動かないstatic剛体になる。</param>
        /// <param name="restitution">反発係数</param>
        /// <param name="friction">摩擦係数</param>
        /// <param name="linear_damp">移動減衰係数</param>
        /// <param name="angular_damp">回転減衰係数</param>
        /// <param name="kinematic">物理演算の影響を受けないKinematic剛体</param>
        /// <param name="group">グループNo.</param>
        /// <param name="mask">衝突するグループNo.</param>
        /// <returns></returns>
        public RigidBody CreateSphere(float radius, Matrix world, float mass = 0, float restitution = 0,
            float friction = 0.5f,
            float linear_damp = 0, float angular_damp = 0, bool kinematic = false,
            CollisionFilterGroups group = CollisionFilterGroups.DefaultFilter,
            CollisionFilterGroups mask = CollisionFilterGroups.AllFilter
            )
        {
            var shape = new SphereShape(radius);
            if (kinematic) mass = 0;
            return CreateShape(shape,world, mass, restitution, friction, linear_damp, angular_damp, kinematic, group,
                mask);
        }

        // 中心軸はZ軸
        public RigidBody CreateCylinder(float radius, float length, Matrix world, float mass = 0, float restitution = 0,
            float friction = 0.5f, float linear_damp = 0, float angular_damp = 0, bool kinematic = false,
            CollisionFilterGroups group = CollisionFilterGroups.DefaultFilter,
            CollisionFilterGroups mask = CollisionFilterGroups.AllFilter
            )
        {
            CollisionShape shape = new CylinderShape(radius, radius, length);
            if (kinematic) mass = 0;
            return CreateShape(shape, world, mass, restitution, friction, linear_damp, angular_damp, kinematic, group,
                mask);
        }

        // 中心軸はZ軸. heightは球の中心間の距離
        public RigidBody CreateCapsule(float radius, float height, Matrix world, float mass = 0, float restitution = 0,
            float friction = 0.5f, float linear_damp = 0, float angular_damp = 0, bool kinematic = false,
            CollisionFilterGroups group = CollisionFilterGroups.DefaultFilter,
            CollisionFilterGroups mask = CollisionFilterGroups.AllFilter
            )
        {
            CollisionShape shape = new CapsuleShape(radius, height);
            if (kinematic) mass = 0;
            return CreateShape(shape, world, mass, restitution, friction, linear_damp, angular_damp, kinematic, group,
                mask);
        }


        // 拘束条件追加
        public void AddPointToPointConstraint(RigidBody body, ref Vector3 pivot)
        {
            dynamicsWorld.AddConstraint(new Point2PointConstraint(body, pivot));
        }

        public void AddPointToPointConstraint(RigidBody bodyA, RigidBody bodyB, ref Vector3 pivotInA,
            ref Vector3 pivotInB)
        {
            dynamicsWorld.AddConstraint(new Point2PointConstraint(bodyA, bodyB, pivotInA, pivotInB));
        }

        /// <summary>
        /// 6軸ジョイントを追加
        /// </summary>
        /// <param name="bodyA">剛体A</param>
        /// <param name="bodyB">剛体B</param>
        /// <param name="frameInA">ジョイントのワールド変換行列(剛体Aローカル座標系)</param>
        /// <param name="frameInB">ジョイントのワールド変換行列(剛体Bローカル座標系)</param>
        /// <param name="c_p1">移動制限1</param>
        /// <param name="c_p2">移動制限2</param>
        /// <param name="c_r1">回転制限1</param>
        /// <param name="c_r2">回転制限2</param>
        /// <param name="stiffness">バネ剛性(平行移動x, y, z, 回転移動x, y, zの順の6要素)</param>
        public void Add6DofSpringConstraint(RigidBody bodyA, RigidBody bodyB, Matrix frameInA, Matrix frameInB,
            Vector3 c_p1, Vector3 c_p2, Vector3 c_r1, Vector3 c_r2, List<float> stiffness)
        {
            Generic6DofSpringConstraint constraint = new Generic6DofSpringConstraint(bodyA, bodyB, frameInA, frameInB,
                true); // 第五引数の効果は謎。どちらでも同じ様に見える……。
            constraint.LinearLowerLimit = new Vector3(c_p1.X, c_p1.Y, c_p1.Z); // 型はベクトルだがベクトル量ではないのでZは反転しない。
            constraint.LinearUpperLimit = new Vector3(c_p2.X, c_p2.Y, c_p2.Z);
            constraint.AngularLowerLimit = new Vector3(c_r1.X, c_r1.Y, c_r1.Z);
            constraint.AngularUpperLimit = new Vector3(c_r2.X, c_r2.Y, c_r2.Z);
            for (int i = 0; i < stiffness.Count(); ++i)
            {
                if (stiffness[i] != 0.0f)
                {
                    constraint.EnableSpring(i, true);
                    constraint.SetStiffness(i, stiffness[i]);
                }
            }
            dynamicsWorld.AddConstraint(constraint);
        }

        // 剛体を移動
        public void MoveRigidBody(RigidBody body, Matrix world)
        {
            body.MotionState.WorldTransform = world;
        }

        // 物理演算の世界の時間を進める
        public void StepSimulation()
        {
            dynamicsWorld.StepSimulation(1.0f/30.0f, 10);
        }

        public void StepSimulation(float tick)
        {
            dynamicsWorld.StepSimulation(tick, 10);
        }

        // 結果のワールド行列を取得
        public Matrix GetWorld(RigidBody body)
        {
            return body.MotionState.WorldTransform;
        }
    }
}