﻿using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using BulletSharp;
using MMF.CG.Model.MMD.Morph;
using MMF.CG.Skinning;
using MMDFileParser.PMXModelParser;
using MMDFileParser.PMXModelParser.JointParam;
using SlimDX;
using IDisposable = System.IDisposable;

namespace MMF.CG.Physics {

	/// <summary>
	/// 物理演算ボーンのマネージャ
	/// </summary>
	public class PhysicsTransformManager:ITransformUpdater,IDisposable
    {

		// 最初に計算しておいてあとで繰り返し使う剛体データ
		class TempRigidBodyData 
        {
            /// <summary>
            /// 初期位置
            /// </summary>
			public readonly Vector3 position;
            /// <summary>
            /// 初期姿勢行列
            /// </summary>
			public readonly Matrix init_matrix;
            /// <summary>
            /// オフセット行列
            /// </summary>
			public readonly Matrix offset_matrix;
            /// <summary>
            /// ボーンインデックス
            /// </summary>
			public readonly int boneIndex;
            /// <summary>
            /// 物理計算タイプ
            /// </summary>
			public readonly PhysicsCalcType physicsCalcType;
            /// <summary>
            /// 剛体形状
            /// </summary>
			public readonly RigidBodyShape shape;

            /// <summary>
            /// 剛体データのテンポラリ
            /// </summary>
            /// <param name="rigidBodyData">剛体データ</param>
			public TempRigidBodyData(RigidBodyData rigidBodyData)
            {
				position = rigidBodyData.Position;
				var r = rigidBodyData.Rotation;
				init_matrix = Matrix.RotationYawPitchRoll(r.Y, r.X, r.Z) * Matrix.Translation(position);
				offset_matrix = Matrix.Invert(init_matrix);
				boneIndex = rigidBodyData.BoneIndex;
				physicsCalcType = rigidBodyData.PhysicsCalcType;
				shape = rigidBodyData.Shape;
			}
		}
        /// <summary>
        /// ボーン
        /// </summary>
		Bone[] bones;
        /// <summary>
        /// 剛体データのテンポラリリスト
        /// </summary>
		List<TempRigidBodyData> tempRigidBodyData_s = new List<TempRigidBodyData>();
		/// <summary>
		/// Bullet物理演算
		/// </summary>
        BulletPhysics bulletPhysics;
        /// <summary>
        /// Bulletの剛体リスト
        /// </summary>
		List<BulletSharp.RigidBody> rigidBodies = new List<BulletSharp.RigidBody>();

        /// <summary>
        /// 剛体を生成する
        /// </summary>
        /// <param name="rigidBodyData_s">剛体データ</param>
        void CreateRigid(List<RigidBodyData> rigidBodyData_s)
        {
            foreach (var r in rigidBodyData_s)
            {
                var tempRigidBodyData = new TempRigidBodyData(r);
                var init_matrix = tempRigidBodyData.init_matrix;
                tempRigidBodyData_s.Add(tempRigidBodyData);
                BulletSharp.RigidBody rigidBody = null;
                ConvexShape convex;
                // 剛体形状ごとに処理する
                switch(r.Shape)
                {
                        
                    // 球体
                    case RigidBodyShape.Sphere:
                        rigidBody = bulletPhysics.CreateSphere(
                                r.Size.X,
                                init_matrix, 
                                r.Mass, 
                                r.Repulsion,
                                r.Friction,
                                r.MoveAttenuation,
                                r.RotationAttenuation,
                                r.PhysicsCalcType == PhysicsCalcType.Static,
                                (BulletSharp.CollisionFilterGroups)(1 << r.RigidBodyGroup),
                                (BulletSharp.CollisionFilterGroups)r.UnCollisionGroupFlag);
                        rigidBodies.Add(rigidBody);
                        break;
                
                    // ボックス
                    case RigidBodyShape.Box:
                        rigidBody = bulletPhysics.CreateBox(
                                r.Size.X,
                                r.Size.Y,
                                r.Size.Z,
                                init_matrix,
                                r.Mass,
                                r.Repulsion,
                                r.Friction,
                                r.MoveAttenuation,
                                r.RotationAttenuation,
                                r.PhysicsCalcType == PhysicsCalcType.Static,
                                (BulletSharp.CollisionFilterGroups)(1 << r.RigidBodyGroup),
                                (BulletSharp.CollisionFilterGroups)r.UnCollisionGroupFlag);
                        rigidBodies.Add(rigidBody);
                        break;
                    // カプセル
                    case RigidBodyShape.Capsule:
                        rigidBody = bulletPhysics.CreateCapsule(
                                    r.Size.X,
                                    r.Size.Y,
                                    init_matrix,
                                    r.Mass,
                                    r.Repulsion,
                                    r.Friction,
                                    r.MoveAttenuation,
                                    r.RotationAttenuation,
                                    r.PhysicsCalcType == PhysicsCalcType.Static,
                                    (BulletSharp.CollisionFilterGroups)(1 << r.RigidBodyGroup),
                                    (BulletSharp.CollisionFilterGroups)r.UnCollisionGroupFlag);
                        rigidBodies.Add(rigidBody);
                        break;
                    default:
                        // TODO: 例外処理、データが壊れている旨を伝える処理を記述する必要有り。
                        break;
                }
            }
        }
        /// <summary>
        /// ジョイント生成
        /// </summary>
        /// <param name="jointData_s">ジョイントデータリスト</param>
		void CreateJoint(List<JointData> jointData_s) 
        {
			foreach (var jointData in jointData_s)
            {
				var joi = (Spring6DofJointParam)jointData.JointParam;
				var a = rigidBodies[joi.RigidBodyAIndex];
				var a_inv = Matrix.Invert(bulletPhysics.GetWorld(a));
				var b = rigidBodies[joi.RigidBodyBIndex];
				var b_inv = Matrix.Invert(bulletPhysics.GetWorld(b));
				var rot = joi.Rotation;
				var pos = joi.Position;
				var world = Matrix.RotationYawPitchRoll(rot.Y, rot.X, rot.Z) * Matrix.Translation(pos.X, pos.Y, pos.Z);
				var stiffness = new List<float>();

                for (int i = 0; i < 3; ++i)
                {
                    stiffness.Add(joi.SpringMoveCoefficient[i]);
                }
                for (int i = 0; i < 3; ++i)
                {
                    stiffness.Add(joi.SpringRotationCoefficient[i]);
                }
                bulletPhysics.Add6DofSpringConstraint(
                    a,
                    b, 
                    world * a_inv,
                    world * b_inv, 
					joi.MoveLimitationMin, 
                    joi.MoveLimitationMax,
                    joi.RotationLimitationMin, 
                    joi.RotationLimitationMax,
                    stiffness);
			}
		}

		/// <summary>
		/// コンストラクタ
		/// </summary>
		/// <param name="bones">ボーンリスト</param>
		/// <param name="rigidBodies">剛体リスト</param>
		public PhysicsTransformManager(Bone[] bones, List<RigidBodyData> rigidBodyData_s, 
            List<JointData> jointData_s)
        {
			this.bones = bones;
			var gravity = new Vector3(0, -9.8f*2.5f, 0);
			bulletPhysics = new BulletPhysics(gravity);
			CreateRigid(rigidBodyData_s);
			CreateJoint(jointData_s);
		}

		/// <summary>
		/// アンマネージなBulletを手動で捨てる
		/// </summary>
		public void Dispose()
        {
            this.tempRigidBodyData_s.Clear();
            this.bulletPhysics.Dispose();
            this.rigidBodies.Clear();
		}

	    /// <summary>
	    /// 物理演算ボーンを更新
	    /// </summary>
        /// <param name="prov">ISkinningProvider</param>
        /// <param name="morph">IMMDMorphManager　モーフ管理クラスのインターフェース</param>
	    bool ITransformUpdater.UpdateTransform(ISkinningProvider prov, IMMDMorphManager morph)
        {
            // 下記の処理では剛体タイプがボーン追従であるもの全てをボーン行列に設定し
            // 物理演算を行った後に剛体タイプが物理演算のものを設定、その後に剛体タイプが
            // 物理＋ボーン位置あわせの物を設定する。設定および計算を行う順番に注意すること。

            // ボーン追従タイプの剛体にボーン行列を設定
			for (int i = 0; i < rigidBodies.Count; ++i) 
            { 
				var t = tempRigidBodyData_s[i];
                // 関連ボーン有りで剛体タイプがボーン追従の場合
				if (t.boneIndex != -1 && t.physicsCalcType == PhysicsCalcType.Static)
                {
                    // ボーン行列を設定処理
                    var bone = bones[t.boneIndex];
					var rigidMat = t.init_matrix*bone.GlobalPose;
					bulletPhysics.MoveRigidBody(rigidBodies[i], rigidMat);
				}
			}
            // 物理演算シミュレーション
			bulletPhysics.StepSimulation();

            // 物理演算の結果に合わせてボーンを設定
			for (int i = 0; i <  rigidBodies.Count; ++i)
            {
				var t = tempRigidBodyData_s[i];
                // 関連ボーンなし時は処理しない
				if (t.boneIndex != -1)
                {
					var bone = bones[t.boneIndex];
					bone.GlobalPose = t.offset_matrix*bulletPhysics.GetWorld(rigidBodies[i]);
				}
			}

            // ボーン位置あわせタイプの剛体の位置移動量にボーンの位置移動量を設定
			for (int i = 0; i < rigidBodies.Count; ++i)
            { 
				var t = tempRigidBodyData_s[i];
                // 関連ボーン有りで剛体タイプが物理＋ボーン位置あわせの場合ボーンの位置移動量を設定
				if (t.boneIndex != -1 && t.physicsCalcType == PhysicsCalcType.BoneAlignment)
                {
					var bone = bones[t.boneIndex];
					var v = new Vector3(bone.GlobalPose.M41, bone.GlobalPose.M42, bone.GlobalPose.M43);	// ボーンの移動量
					var p = new Vector3(t.init_matrix.M41, t.init_matrix.M42, t.init_matrix.M43) + v;
					var m = bulletPhysics.GetWorld(rigidBodies[i]);
					m.M41 = p.X; m.M42 = p.Y; m.M43 = p.Z;
					bulletPhysics.MoveRigidBody(rigidBodies[i], m);
				}
			}
            return false;
	    }
	}
}
