﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using MMDFileParser.PMXModelParser;
using SlimDX;

namespace MMF.CG.Skinning
{
    /// <summary>
    ///     ボーンクラス
    /// </summary>
    public class Bone
    {
        private readonly ISkinningProvider skinning;

        /// <summary>
        ///     ボーンのインデックス
        /// </summary>
        public int BoneIndex;

        /// <summary>
        ///     ボーン名
        /// </summary>
        public string BoneName;

        /// <summary>
        ///     子ボーン
        /// </summary>
        public List<Bone> Children = new List<Bone>();

        public Vector3 DefaultLocalX;
        public Vector3 DefaultLocalY;
        public Vector3 DefaultLocalZ;

        /// <summary>
        ///     親ボーン
        /// </summary>
        public Bone Parent;

        /// <summary>
        ///     元々のローカル位置ベクタ
        /// </summary>
        public Vector3 Position;

        /// <summary>
        ///     平行移動位置ベクタ
        /// </summary>
        public Vector3 Translation;

        public bool isLocalAxis;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        /// <param name="bones"></param>
        /// <param name="index"></param>
        /// <param name="layer"></param>
        /// <param name="skinning"></param>
        public Bone(List<BoneData> bones, int index, int layer, ISkinningProvider skinning)
        {
            this.skinning = skinning;
            BoneData me = bones[index]; //このボーン
            skinning.Bone[index] = this;
            BoneIndex = index;
            Position = me.Position;
            BoneName = me.BoneName;
            isLocalAxis = me.isLocalAxis;
            PhysicsOrder = me.transformAfterPhysics ? PhysicsOrder.After : PhysicsOrder.Before;
            Layer = layer;
            if (isLocalAxis)
            {
                DefaultLocalX = me.DimentionXDirectionVector;
                DefaultLocalY = Vector3.Cross(me.DimentionZDirectionVector, DefaultLocalX);
                DefaultLocalZ = Vector3.Cross(DefaultLocalX, DefaultLocalY);
            }
            if (me.isIK) //IKボーンの場合
            {
                skinning.IkBone.Add(this);
                isIK = true;
                RotationLimited = me.IKLimitedRadian;
                targetBoneIndex = me.IKTargetBoneIndex;
                Iterator = me.IKLoopNumber;
                foreach (IKLink ikLink in me.ikLinks)
                {
                    IkLink link = new IkLink(skinning, ikLink.LinkBoneIndex);
                    Vector3 maxVec = ikLink.MinimumRadian;
                    Vector3 minVec = ikLink.MaximumRadian;
                    //minとmaxを正しく読み込む
                    link.minRot = new Vector3(Math.Min(maxVec.X, minVec.X), Math.Min(maxVec.Y, minVec.Y),
                        Math.Min(maxVec.Z, minVec.Z));
                    link.maxRot = new Vector3(Math.Max(maxVec.X, minVec.X), Math.Max(maxVec.Y, minVec.Y),
                        Math.Max(maxVec.Z, minVec.Z));
                    link.isLimited = ikLink.isRotateLimited;
                    ikLinks.Add(link);
                }
            }
            for (int i = 0; i < bones.Count; i++)
            {
                BoneData bone = bones[i];
                if (bone.ParentBoneIndex == index)
                {
                    Bone child = new Bone(bones, i, layer + 1, skinning);
                    AddChild(child);
                }
            }
        }

        /// <summary>
        ///     回転行列
        /// </summary>
        public Quaternion Rotation
        {
            get { return rotation; }
            set
            {
                rotation = value;
                rotation.Normalize();
            }
        }

        /// <summary>
        ///     変形階層
        /// </summary>
        public int Layer { get; private set; }

        /// <summary>
        ///     物理順
        /// </summary>
        public PhysicsOrder PhysicsOrder { get; private set; }


        public Vector3 LocalX
        {
            get { return Vector3.TransformNormal(DefaultLocalX, Matrix.RotationQuaternion(Rotation)); }
        }

        public Vector3 LocalY
        {
            get { return Vector3.TransformNormal(DefaultLocalY, Matrix.RotationQuaternion(Rotation)); }
        }

        public Vector3 LocalZ
        {
            get { return Vector3.TransformNormal(DefaultLocalZ, Matrix.RotationQuaternion(Rotation)); }
        }


        /// <summary>
        ///     ローカルポーズ
        /// </summary>
        public Matrix LocalPose { get; set; }

        /// <summary>
        ///     グローバルポーズ
        /// </summary>
        public Matrix GlobalPose { get;  set; }

        #region IKに関する項目

        private readonly int targetBoneIndex;

        /// <summary>
        ///     演算処理の繰り返し数
        /// </summary>
        public int Iterator;

        /// <summary>
        ///     １回あたりの演算の制限回転量
        /// </summary>
        public float RotationLimited;

        /// <summary>
        ///     IKリンクのリスト
        /// </summary>
        public List<IkLink> ikLinks = new List<IkLink>();

        /// <summary>
        ///     このボーンがIKボーンかどうか
        /// </summary>
        public bool isIK = false;

        private Quaternion rotation;

        public Bone IkTargetBone
        {
            get { return skinning.Bone[targetBoneIndex]; }
        }

        #endregion

        /// <summary>
        ///     子関節を追加します。
        /// </summary>
        /// <param name="child">追加する子関節</param>
        public void AddChild(Bone child)
        {
            Children.Add(child);
            child.Parent = this;
        }

        public void UpdateGrobalPose()
        {
            LocalPose = Matrix.Translation(-Position)*Matrix.RotationQuaternion(Rotation)*
                        Matrix.Translation(Translation)*Matrix.Translation(Position);
            GlobalPose = LocalPose;
            if (Parent != null) //この要素がルートではないとき
            {
                GlobalPose *= Parent.GlobalPose;
            }
            Parallel.ForEach(Children, (item) => item.UpdateGrobalPose());
        }
    }
}