﻿using System;
using System.Collections.Generic;
using MMF.CG.Model.MMD.Morph;
using MMF.CG.Utility;
using SlimDX;

namespace MMF.CG.Skinning
{
    /// <summary>
    ///     CCDIKのIKProvider
    /// </summary>
    public class CCDIK : IIKProvider
    {
        /// <summary>
        ///     IK解決
        /// </summary>
        /// <param name="prov"></param>
        /// <param name="morph"></param>
        bool ITransformUpdater.UpdateTransform(ISkinningProvider prov, IMMDMorphManager morph)
        {
            //すべてのIKボーンに対して処理
            foreach (Bone ikBone in prov.IkBone
                )
            {
                foreach (var link in ikBone.ikLinks)
                {
                    link.loopCount = 0;
                }
                /*
                 *CCD-IKの仕組み
                 *-◎-◎-◎-●
                 * ↑関節      ▲←IKターゲット(MMD内ではIKBoneと呼ぶ)。関節を近づけたい目標
                 * (関節内の◎にあたる、IKで曲げる途中の関節をIKリンク、●にあたる関節の終点をエフェクタ(MMD内ではターゲットボーン)
                 * とする。
                 * 1)まず、◎-●の方向ベクタを求める。ベクトルA
                 * 2)次に、◎-▲の方向ベクタを求める。ベクトルB
                 * 3)∠●◎▲における最短回転角度を求める。arccos(A・B)よりラジアン単位で求まる
                 * 4)２つの方向ベクトルの外積から回転軸を求める
                 */
                //エフェクタのボーンを取得
                Bone effector = ikBone.IkTargetBone;
                Vector3 TargetGlobalPos = Vector3.TransformCoordinate(ikBone.Position, ikBone.GlobalPose);
                for (int it = 0; it < ikBone.Iterator; it++) //Ikボーンごとに定められた繰り返し数だけ演算を繰り返す。
                {

                    for (int linkIndex = 0; linkIndex < ikBone.ikLinks.Count; linkIndex++) //それぞれのリンクに対して処理をする。
                    {
                        IkLink ikLink = ikBone.ikLinks[linkIndex];
                        Matrix ToLinkLocal = Matrix.Invert(ikLink.ikLinkBone.GlobalPose);
                        //ターゲットの座標は以下のループ中で変更されないので先に求めておく ▲
                        Vector3 targetPos;Vector3.TransformCoordinate(ref TargetGlobalPos, ref ToLinkLocal, out targetPos);
                        Vector3 effectorPos = Vector3.TransformCoordinate(effector.Position,effector.GlobalPose*ToLinkLocal); //●
                        Vector3 linkPos =ikLink.ikLinkBone.Position;//◎

                        Vector3 link2Effector = Vector3.Normalize(effectorPos - linkPos);
                        Vector3 link2Target = Vector3.Normalize(targetPos - linkPos);
                        //回転角度を求める
                        float dot = Vector3.Dot(link2Effector, link2Target);
                        if (dot > 1f)
                        {
                            dot = 1f;
                        }
                        float rotationAngle = (float) Math.Acos(dot);
                        ////角度制限を適用する
                        if (rotationAngle > ikBone.RotationLimited) rotationAngle = ikBone.RotationLimited;
                        if (rotationAngle < -ikBone.RotationLimited) rotationAngle = -ikBone.RotationLimited;

                        //回転軸を求める
                        Vector3 rotationAxis = Vector3.Cross(link2Effector, link2Target);
                        if (!float.IsNaN(rotationAngle) && rotationAngle > 1.0e-3f)
                        {
                            ikLink.loopCount++;
                            //軸を中心として回転する行列を作成する。
                            Quaternion rotation = Quaternion.RotationAxis(rotationAxis, rotationAngle);
                            rotation.Normalize();
                            ikLink.ikLinkBone.Rotation = rotation*ikLink.ikLinkBone.Rotation;

                            #region 回転量制限

                            //回転量制限
                            if (ikLink.isLimited)
                            {
                                float xRotation, yRotation, zRotation; //X,Y,Z軸回転量に変換する。
                                int type = 0; //分解パターン
                                if (
                                    !CGHelper.FactoringQuaternionXYZ(ikLink.ikLinkBone.Rotation, out xRotation,
                                        out yRotation, out zRotation))
                                {
                                    if (
                                        !CGHelper.FactoringQuaternionYZX(ikLink.ikLinkBone.Rotation, out yRotation,
                                            out zRotation, out xRotation))
                                    {
                                        CGHelper.FactoringQuaternionZXY(ikLink.ikLinkBone.Rotation, out zRotation,
                                            out xRotation, out yRotation);
                                        type = 2;
                                    }
                                    else
                                        type = 1;
                                }
                                else type = 0;
                                xRotation = CGHelper.Clamp(xRotation, ikLink.minRot.X, ikLink.maxRot.X);
                                yRotation = CGHelper.Clamp(yRotation, ikLink.minRot.Y, ikLink.maxRot.Y);
                                zRotation = CGHelper.Clamp(zRotation, ikLink.minRot.Z, ikLink.maxRot.Z);
                                switch (type)
                                {
                                    case 0:
                                        ikLink.ikLinkBone.Rotation =
                                            Quaternion.RotationMatrix(Matrix.RotationX(xRotation) *
                                                                      Matrix.RotationY(yRotation) *
                                                                      Matrix.RotationZ(zRotation));
                                        break;
                                    case 1:
                                        ikLink.ikLinkBone.Rotation =
                                            Quaternion.RotationMatrix(Matrix.RotationY(yRotation) *
                                                                      Matrix.RotationZ(zRotation) *
                                                                      Matrix.RotationX(xRotation));
                                        break;
                                    case 2:
                                        ikLink.ikLinkBone.Rotation = Quaternion.RotationYawPitchRoll(xRotation,
                                            xRotation, zRotation);
                                        break;
                                }
                            }

                            #endregion

                            ikLink.ikLinkBone.UpdateGrobalPose();
                        }
                    }
                }
            }
            return false;
        }
    }
}