﻿using System;
using System.Collections.Generic;
using System.Threading.Tasks;
using MMF.CG.Model.MMD.Morph;
using MMDFileParser.PMXModelParser;
using SlimDX;
using SlimDX.Direct3D11;

namespace MMF.CG.Skinning
{
	public class Skeleton : ISkinningProvider
    {
        /// <summary>
        ///     ボーンそれぞれのグローバルポーズ
        /// </summary>
        public Matrix[] GlobalBonePose;

        /// <summary>
        ///     Ikを処理するインターフェース
        /// </summary>
        public IIKProvider IkProvider;

        public List<ITransformUpdater> KinematicsProviders { get; private set; }

        /// <summary>
        ///     ボーンのルート(ルートが2つ以上の場合があり)
        /// </summary>
        public List<Bone> RootBone = new List<Bone>();

        public Skeleton(ModelData model)
        {
            IkProvider = new CCDIK();
            //ボーンの数だけ初期化
            GlobalBonePose = new Matrix[model.BoneList.BoneCount];
            Bone = new Bone[model.BoneList.BoneCount];
            IkBone = new List<Bone>();
            //ボーンを読み込む
            LoadBones(model);
            BoneDictionary=new Dictionary<string, Bone>();
            foreach (var bone in Bone)
            {
                BoneDictionary.Add(bone.BoneName,bone);
            }
            KinematicsProviders=new List<ITransformUpdater>();
            KinematicsProviders.Add(IkProvider);
        }


        /// <summary>
        ///     ボーン
        /// </summary>
        public Bone[] Bone { get; set; }

	    public Dictionary<string, Bone> BoneDictionary { get; private set; }

	    /// <summary>
        ///     Ikボーンのリスト
        /// </summary>
        public List<Bone> IkBone { get; set; }

        /// <summary>
        ///     エフェクトに、現在のボーン情報を渡します。
        /// </summary>
        /// <param name="effect"></param>
        public void ApplyEffect(Effect effect)
        {
            effect.GetVariableBySemantic("BONETRANS").AsMatrix().SetMatrixArray(GlobalBonePose);
        }

	    /// <summary>
	    ///     スキニング全体を更新します
	    /// </summary>
	    /// <param name="morph"></param>
	    public virtual void UpdateSkinning(IMMDMorphManager morph)
        {
            ResetAllBoneTransform();
            UpdateGlobal();
            foreach (ITransformUpdater kinematicsProvider in KinematicsProviders)
            {
                if (kinematicsProvider.UpdateTransform(this,morph))
                {
                    UpdateGlobal();
                    ResetAllBoneTransform();
                }
            }
            for (int i = 0; i < Bone.Length; i++)
            {
                GlobalBonePose[Bone[i].BoneIndex] = Bone[i].GlobalPose;
            }
        }

        /// <summary>
        ///     すべての回転・移動を元に戻します。
        /// </summary>
        public void ResetAllBoneTransform()
        {
            foreach (Bone item in Bone)
            {
                item.Rotation = Quaternion.Identity;
                item.Translation = Vector3.Zero;
            }
        }

        private void LoadBones(ModelData model)
        {
            for (int i = 0; i < model.BoneList.BoneCount; i++)
            {
                if (model.BoneList.Bones[i].ParentBoneIndex == model.GetBoneNoReferenceCode()
                    ||
                    model.BoneList.Bones[i].ParentBoneIndex == -1)
                {
                    RootBone.Add(new Bone(model.BoneList.Bones, i, 0, this));
                }
            }
            BoneComparer comparer = new BoneComparer(model.BoneList.Bones.Count);
            IkBone.Sort(comparer);
            RootBone.Sort(comparer);
        }

        /// <summary>
        ///     現在の回転行列に基づき、ルートボーンからグローバルポーズを再計算します。
        /// </summary>
        protected void UpdateGlobal()
        {
            Parallel.ForEach(RootBone, (item) => item.UpdateGrobalPose());
        }

		public virtual void Dispose() {}
    }
}